/*
 * $RCSfile: ethtap.c,v $
 *
 * Copyright (c) 1997  Arizona Board of Regents
 *
 *
 * This code was written for use with the x-kernel for filtering in
 * user-space all network traffic.  It may have further applications.
 *
 *
 * ETHTAP is a kernel module which can be used to divert all traffic
 * to/from a lower level nic to a user-space process.  Due to the
 * nature of this code, both this file, and the lower level nic driver
 * must be compiled as modules.
 *
 * To use ETHTAP load it prior to loading the nic driver to be
 * affected.  If you load more than one nic driver only the first is
 * affected by ETHTAP.
 *
 * ETHTAP implement two psuedo devices.
 *
 * /dev/htap - high tap
 *	read()  -> reads outgoing ethernet frames
 *	write() <- writes ethernet frames to kernel stack
 *
 * /dev/ltap - low tap
 *	read()  -> reads incomming ethernet frames
 *	write() <- writes ethernet frames to nic driver
 *
 * Both the LTAP and HTAP devices can be placed in either ACTIVE or
 * PASSIVE mode, with PASSIVE the default after an open().  In PASSIVE
 * mode copies of all frames are made to allow read()s, and all
 * write()s are silently discarded.  In ACTIVE mode all frames are
 * queued for subsequent read()s and traffic will not flow through the
 * network stack.
 *
 * For filtering packets, set both LTAP and HTAP to ACTIVE mode with
 * an ioctl() on each.  To establish the outgoing path, read from HTAP
 * and write verbatum to LTAP.  To establish the incomming, path read
 * from LTAP and write verbatum to HTAP.  Any packet filtering, or
 * inspection should take place between to read()/writes().
 *
 * In order to minimize the read()/write() syscalls used a special
 * scheme is implemented for packing multiple ethernet frames into a
 * single buffer.  All read()s will return a buffer of the form.
 *
 * <frames>byte <length>word <frame>? <length>word <frame>? ...
 *
 * Where the first byte indicated the number of frames read.  The next
 * two bytes indicates the length of the following frame, and this
 * continues FRAMES times.
 *
 * All data written must follow the same format.  For example to write
 * a single packet, it should be of the form:
 *
 * 0x01 <length>word <frame>?
 *
 * Although this adds three bytes for single frame writes, experience
 * shows read()s from LTAP will often contain upto 16 frames.
 *
 *
 * $Log: ethtap.c,v $
 * Revision 1.2  1997/05/09 22:51:12  mjk
 * now use dev_kfree_skb where needed.
 *
 * Revision 1.1  1997/04/23 19:15:38  mjk
 * Initial revision
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/interrupt.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <linux/ethtap.h>


/* 
 *  Defines
 */

#define ETHTAP_MAX_QSIZE	300
#define ETHTAP_MAX_READPKTS	16




/* 
 * Types
 */


typedef struct {
  char			*name;	/* string used for error messages 	*/
  int			mode;	/* state of the tap 			*/
  struct sk_buff_head	rxq;	/* rx frame list			*/
  int			rxsize;	/* 	size of frame list 		*/
  long			rxlost;	/* 	num packets dropped 		*/
  long			rxread;	/* 	frames read 			*/
  long			rxwrite;/* 	frames written			*/
  struct sk_buff_head	txq;	/* tx frame list			*/
  int			txsize;	/* 	size of frame list 		*/
  long			txlost;	/* 	num packets dropped 		*/
  long			txread;	/* 	frames read 			*/
  long			txwrite;/* 	frames written			*/
  struct sk_buff	*lskb;	/* last skb read 			*/
				/* write method 			*/
  int			(*write_skb)(const char *, int);
} tap_t;




/* 
 * Prototypes 
 */


static int	get_info(char *, char **, off_t, int, int);

static int	bogus_hard_start_xmit(struct sk_buff *, struct device *);
static int	bogus_register_netdev(struct device *);
static void	bogus_unregister_netdev(struct device *);
static void	bogus_netif_rx(struct sk_buff *);
static void	bogus_dev_tint(struct device *);


static void	tap_purge_rx(tap_t *, int, int);
static void	tap_purge_tx(tap_t *, int, int);
static void	tap_init(tap_t *, const char *, int (*)(const char *, int));

static int	high_read(struct inode *, struct file *, char *, int);
static int	high_write(struct inode *, struct file *, const char *, int);
static int	high_ioctl(struct inode *, struct file *, unsigned,
			   unsigned long);
static int	high_open(struct inode *, struct file *);
static void	high_release(struct inode *, struct file *);
static int	high_write_skb(const char *, int);

static int	low_read(struct inode *, struct file *, char *, int);
static int	low_write(struct inode *, struct file *, const char *, int);
static int	low_ioctl(struct inode *, struct file *, unsigned,
			   unsigned long);
static int	low_open(struct inode *, struct file *);
static void	low_release(struct inode *, struct file *);
static int	low_write_skb(const char *, int);
static int	low_do_hard_start_xmit(void);


static int	shared_write(tap_t *, struct inode *, struct file *,
			     const char *, int);
static int	shared_ioctl(tap_t *, struct inode *, struct file *,
			     unsigned, unsigned long);
static int	shared_open(tap_t *, struct inode *, struct file *,
			    int);
static void	shared_release(tap_t *, struct inode *, struct file *);




/* 
 *  Globals
 */

static int		activep	= 0; /* tap active predicate 		*/
static struct device	*lld	= NULL; /* lower level device 		*/
static tap_t	lowtap;	 /* LOW  traffic tap 			*/
static tap_t	hightap; /* HIGH traffic tap 			*/

				/* lower level device tx method 	*/
static int		(*hard_start_xmit)(struct sk_buff *, struct device *);




				/* Exported symbol table 		*/
static struct symbol_table	ethtap_symbols = {
#include <linux/symtab_begin.h>
  Xalias(bogus_register_netdev, register_netdev),
  Xalias(bogus_unregister_netdev, unregister_netdev),
  Xalias(bogus_netif_rx, netif_rx),
  Xalias(bogus_dev_tint, dev_tint),
#include <linux/symtab_end.h>
};

				/* /proc/net/ethtap interface 		*/
static struct proc_dir_entry	ethtap_proc = {
  PROC_NET_ETHTAP,		/* low_ino 				*/
  6,				/* namelen 				*/
  "ethtap",			/* name 				*/
  S_IFREG | S_IRUGO,		/* mode 				*/
  1,				/* nlink 				*/
  0,				/* uid 					*/
  0,				/* gid 					*/
  0,				/* size 				*/
  &proc_net_inode_operations,	/* ops 					*/
  get_info		/* get_info() 				*/
};


static struct file_operations	high_fops = {
  NULL,				/* lseek 				*/
  high_read,			/* read 				*/
  high_write,			/* write 				*/
  NULL,				/* readdir 				*/
  NULL,				/* select 				*/
  high_ioctl,			/* ioctl 				*/
  NULL,				/* mmap 				*/
  high_open,			/* open 				*/
  high_release,			/* release 				*/
  NULL,				/* fsync 				*/
  NULL,				/* fasync 				*/
  NULL,				/* check_media_change 			*/
  NULL				/* revalidate 				*/
};


static struct file_operations	low_fops = {
  NULL,				/* lseek 				*/
  low_read,			/* read 				*/
  low_write,			/* write 				*/
  NULL,				/* readdir 				*/
  NULL,				/* select 				*/
  low_ioctl,			/* ioctl 				*/
  NULL,				/* mmap 				*/
  low_open,			/* open 				*/
  low_release,			/* release 				*/
  NULL,				/* fsync 				*/
  NULL,				/* fasync 				*/
  NULL,				/* check_media_change 			*/
  NULL				/* revalidate 				*/
};





/* 
 * Module Hooks
 */

int
init_module(void)
{
  TAPDEBUG("init_module\n");
				/* Register replacement kernel
                                   functions */
  register_symtab(&ethtap_symbols);

				/* register /proc/net/ethtap 		*/
  proc_net_register(&ethtap_proc);

				/* Register /dev/htap 			*/
  if ( register_chrdev(ETHTAP_HIGH_MAJOR, "htap", &high_fops) ) {
    printk("ethtap: unable to get major %d\n", ETHTAP_HIGH_MAJOR);
    return -EIO;
  }

				/* Register /dev/ltap 			*/
  if ( register_chrdev(ETHTAP_LOW_MAJOR, "ltap", &low_fops) ) {
    printk("ethtap: unable to get major %d\n", ETHTAP_LOW_MAJOR);
    return -EIO;
  }
  

  return 0;
} /* init_module */


void
cleanup_module(void)
{
  TAPDEBUG("cleanup_module\n");
  
  proc_net_unregister(PROC_NET_ETHTAP);
  unregister_chrdev(ETHTAP_HIGH_MAJOR, "htap");
  unregister_chrdev(ETHTAP_LOW_MAJOR, "ltap");
} /* cleanup_module */




/* 
 *  Replacement Kernel functions
 */


static int
bogus_register_netdev(struct device *dev)
{
  int			retval;

  TAPDEBUG("bogus_register_netdev\n");
  
				/* If we are already tapping a device
				   pass this own through, otherwise
				   capture this one. */
  if ( activep )
    return register_netdev(dev);
  
  activep = 1;
  lld	  = dev;

				/* Initialize HIGH and LOW taps 	*/
  tap_init(&lowtap,  "/dev/ltap", low_write_skb);
  tap_init(&hightap, "/dev/htap", high_write_skb);

				/* Replace tx method and register 	*/
  retval = register_netdev(dev);
  hard_start_xmit      = dev->hard_start_xmit;
  dev->hard_start_xmit = bogus_hard_start_xmit;
  
  printk("ethtap: registered device '%s'\n", dev->name);
  return retval;
} /* bogus_register_netdev */


static void
bogus_unregister_netdev(struct device *dev)
{
  TAPDEBUG("bogus_unregister_netdev\n");
  
  if ( lld != dev ) {		/* non-tapped device 			*/
    return unregister_netdev(dev);
  }

  tap_purge_rx(&lowtap,  ETHTAP_MAX_QSIZE, FREE_READ);
  tap_purge_tx(&lowtap,  ETHTAP_MAX_QSIZE, FREE_WRITE);
  tap_purge_rx(&hightap, ETHTAP_MAX_QSIZE, FREE_WRITE);
  
  printk("ethtap: unregistering device '%s'\n", dev->name);
  unregister_netdev(dev);
  activep = 0;
  lld     = NULL;
} /* bogus_unregister_netdev */


				/* Redirects incomming ethernet frames
                                   as needed */
static void
bogus_netif_rx(struct sk_buff *skb)
{
  struct sk_buff	*skb2;
  int			flags;

  if ( lowtap.mode == ETHTAP_CLOSED_MODE ) {
    return netif_rx(skb);
  }

  switch ( lowtap.mode ) {
    
    case ETHTAP_CLOSED_MODE:
      netif_rx(skb);
      break;
      
    case ETHTAP_ACTIVE_MODE:
      skb->free = 1;		/* to be deleted by read() 		*/
      save_flags(flags); cli();	/* -lock- 				*/
      if ( lowtap.rxsize > ETHTAP_MAX_QSIZE ) {
	printk("%s: queue overflow\n", lowtap.name);
	tap_purge_rx(&lowtap, ETHTAP_MAX_QSIZE/10, FREE_READ);
      }
      skb_queue_tail(&lowtap.rxq, skb);
      lowtap.rxsize++;
      restore_flags(flags);	/* -unlock- 				*/
      break;
      
    case ETHTAP_PASSIVE_MODE:
      if ( (skb2 = skb_clone(skb, GFP_ATOMIC)) ) {
	save_flags(flags); cli(); /* -lock- 				*/
	if ( lowtap.rxsize > ETHTAP_MAX_QSIZE ) {
	  printk("%s: queue overflow\n", lowtap.name);
	  tap_purge_rx(&lowtap, ETHTAP_MAX_QSIZE/10, FREE_READ);
	}
	skb_queue_tail(&lowtap.rxq, skb2);
	lowtap.rxsize++;
	restore_flags(flags);	/* -unlock- 				*/
      }
      netif_rx(skb);		/* pass up original skb 		*/
      break;
      
    default:
      ;
  }
  
} /* bogus_netif_rx */


static void
bogus_dev_tint(struct device *dev) 
{

  if ( lld != dev ) {		/* non-tapped device 			*/
    return dev_tint(dev);
  }

				/* For Active mode try to txmit
                                   another frame, for passive just
                                   pass the tx empty interrupt back
                                   up. */
  if ( lowtap.mode == ETHTAP_ACTIVE_MODE ) {
    low_do_hard_start_xmit();
  }
  else {
    dev_tint(dev);
  }
} /* bogus_dev_tint */




/* 
 *  Network Device Methods
 */


static int
get_info(char *buf, char **start, off_t offset, int length, int dummy)
{
  int			len	= 0;

  if ( activep ) {
    len += sprintf(buf+len, "(%s)\n", lld->name);
    len += sprintf(buf+len,
	   "    |    size    |    read    |    write   |    lost    | mode\n");
    len += sprintf(buf+len, "htap|%12d|%12d|%12d|%12d|%d\n",
		   hightap.rxsize, hightap.rxread, hightap.rxwrite,
		   hightap.rxlost, hightap.mode);
    len += sprintf(buf+len, "ltap|%12d|%12d|%12d|%12d|%d\n",
		   lowtap.rxsize, lowtap.rxread, lowtap.rxwrite,
		   lowtap.rxlost, lowtap.mode);
    len += sprintf(buf+len, "  tx|%12d|%12d|%12d|%12d|\n",
		   lowtap.txsize, lowtap.txread, lowtap.txwrite,
		   lowtap.txlost);
  }
  else {
    len += sprintf(buf, "inactive\n");
  }

  if ( offset >= len ) {
    *start = buf;
    return 0;
  }
  
  *start = buf + offset;
  len -= offset;
  if ( len > length )
    len = length;

  return len;
} /* get_info */


static int
bogus_hard_start_xmit(struct sk_buff *skb, struct device *dev) 
{
  int			retval = 0; /* assume success 			*/
  struct sk_buff	*skb2;
  int			flags;

  switch ( hightap.mode ) {

    case ETHTAP_CLOSED_MODE:
      retval = hard_start_xmit(skb, dev);
      break;

    case ETHTAP_ACTIVE_MODE:
      save_flags(flags); cli();	/* -lock- 				*/
				/* on overflow nuke the oldest 10% of
                                   the queue */
      if ( hightap.rxsize > ETHTAP_MAX_QSIZE ) {
	printk("%s: queue overflow\n", hightap.name);
	tap_purge_rx(&hightap, ETHTAP_MAX_QSIZE/10, FREE_WRITE);
      }
				/* add frame to high list 		*/
      skb_queue_tail(&hightap.rxq, skb);
      hightap.rxsize++;
      restore_flags(flags);	/* -unlock- 				*/

      dev_tint(dev);		/* fake tx empty interrupt 		*/
      
      TAPDEBUG("%s: queued %ld bytes\n", hightap.name, skb->len);
      break;

    case ETHTAP_PASSIVE_MODE:
      if ( (skb2 = skb_clone(skb, GFP_ATOMIC)) ) {
	save_flags(flags); cli(); /* -lock- 				*/
	if ( hightap.rxsize > ETHTAP_MAX_QSIZE ) {
	  printk("%s: queue overflow\n", hightap.name);
	  tap_purge_rx(&hightap, ETHTAP_MAX_QSIZE/10, FREE_WRITE);
	}
	skb_queue_tail(&hightap.rxq, skb2);
	hightap.rxsize++;
	restore_flags(flags);	/* -unlock- 				*/
	TAPDEBUG("%s: queued %ld bytes\n", hightap.name, skb->len);
      }
      retval = hard_start_xmit(skb, dev);
      break;

    default:
      ;
  }

  return retval;
} /* bogus_hard_start_xmit */



/* 
 * /dev/htap (HIGH)  device methods
 */


static int
high_write_skb(const char *buf, int count)
{
  struct sk_buff	*skb  = dev_alloc_skb(count + 2);
  int			flags;

  if ( skb ) {
    skb_reserve(skb, 2);
    memcpy_fromfs(skb_put(skb, count), buf, count);
    skb->dev	  = lld;	/* destination net device 		*/
    skb->protocol = eth_type_trans(skb, lld);

    /* There is no flow control at this point.  Since real ethernet
       packets are dropped to manage the queue size there is no need
       to do it here.  In addition this provides a reliable connection
       to the kernel which insures no xkernel cycles are wasted
       forwarding packets that could be dropped.. */
    save_flags(flags); cli();	/* -lock- 				*/
    netif_rx(skb);
    restore_flags(flags);	/* -unlock- 				*/
    return count;
  }
  
  return 0;
} /* high_write_skb */


				/* The rest of these simply call the
                                   shared method with the HIGH tap as
                                   the first argument. */

static int
high_open(struct inode *inode, struct file *file)
{
  return shared_open(&hightap, inode, file, FREE_WRITE);
} /* high_open */

static void
high_release(struct inode *inode, struct file *file)
{
  return shared_release(&hightap, inode, file);
} /* high_release */

static int
high_ioctl(struct inode *inode, struct file *file, unsigned cmd,
	   unsigned long arg)
{
  return shared_ioctl(&hightap, inode, file, cmd, arg);
} /* high_ioctl */

static int
high_read(struct inode *inode, struct file *file, char *buf, int count)
{
  struct sk_buff	*skblist[ETHTAP_MAX_READPKTS];
  unsigned		minor	= MINOR(inode->i_rdev);
  int			avail	= count - 1;	/* reserve a byte 	*/
  char			*p	= &buf[1];	/* reserved byte 	*/
  unsigned char		pkts;
  struct sk_buff	*skb;
  short			len;
  int			i, j;
  int			flags;
  tap_t		*tap = &hightap;

  if ( minor )
    return -ENODEV;

  /* All reads are done non-blocking.  Collect as many packets as you
   * can and store them into the userspace buffer in the form
   * [<pkts>b<len>2<pkt>?<len>2<pkt>?...].  As we collect packet we
   * can either run out of slots in the skblist or run out of room in
   * the buffer.  Running out of buffer space is detected late so a
   * static pointer is used to keep track of any remaining buffer
   * between reads.  Closing the device will free the space.
   *
   * With interrupts disabled suck up as many packets as possible and
   * then with interrupts enabled fill the userspace buffer.
   */
  i = 0;			/* goto slot zero			*/
  save_flags(flags); cli();	/* -lock- 				*/
  if ( tap->lskb ) {		/* collect extra from last read 	*/
    skblist[i++] = tap->lskb;
  }
  while ( i < ETHTAP_MAX_READPKTS ) {
    tap->lskb = skb_dequeue(&tap->rxq);
    if ( tap->lskb ) {
      tap->rxsize--;
      if (avail > (tap->lskb->len + sizeof(struct ethhdr) + sizeof len)) {
	avail -= tap->lskb->len + sizeof(struct ethhdr) + sizeof len;
	skblist[i] = tap->lskb;	/* hold onto the new packet		*/
	tap->lskb  = NULL;	/* reset the static packet		*/
	i++;			/* next packet slot 			*/
      } else break;
    } else break;
  }
  restore_flags(flags);		/* -unlock- 				*/

  for (j=0; j<i; j++) {
    skb = skblist[j];

    len = (ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN;

    memcpy_tofs(p, &len, sizeof len);
    p += sizeof len;
    memcpy_tofs(p, skb->data, len);
    p += len;

    dev_kfree_skb(skb, FREE_WRITE);
  }
  
				/* fill in total num of pkts stored in
				   the buffer */
  pkts = i;
  memcpy_tofs(buf, &pkts, sizeof pkts);
  tap->rxread += pkts;		/* update statistics 			*/
  
#ifdef CONFIG_ETHTAP_DEBUG
  if ( j )
    TAPDEBUG("%s: read %d pkts\n", tap->name, j);
#endif  

  return pkts ? count - avail : 0;
} /* high_read */


static int
high_write(struct inode *inode, struct file *file, const char *buf, int count) 
{
  return shared_write(&hightap, inode, file, buf, count);
} /* high_write */




/* 
 *  /dev/ltap (LOW) Device Methods
 */

static int
low_write_skb(const char *buf, int count)
{
  struct sk_buff	*skb = alloc_skb(count, GFP_ATOMIC);
  int			flags;

  if ( skb ) {
    memcpy_fromfs(skb_put(skb, count), buf, count);
    skb->free = 1;		/* driver frees skb 			*/

    save_flags(flags); cli();	/* -lock- 				*/
    if ( lowtap.txsize > ETHTAP_MAX_QSIZE ) {
      printk("%s: tx queue overflow\n", lowtap.name);
      tap_purge_tx(&lowtap, ETHTAP_MAX_QSIZE/10, FREE_READ);
    }
    skb_queue_tail(&lowtap.txq, skb);
    lowtap.txsize++;
    lowtap.txwrite++;
    restore_flags(flags);	/* -unlock- 				*/
    
				/* Try now to transmit 			*/
    low_do_hard_start_xmit();
    
    return count;
  }
  
  return 0;
} /* low_write_skb */


				/* The rest of these simply call the
                                   shared method with the HIGH tap as
                                   the first argument. */


static int
low_open(struct inode *inode, struct file *file)
{
  return shared_open(&lowtap, inode, file, FREE_READ);
} /* low_open */

static void
low_release(struct inode *inode, struct file *file)
{
  return shared_release(&lowtap, inode, file);
} /* low_release */

static int
low_ioctl(struct inode *inode, struct file *file, unsigned cmd,
	   unsigned long arg)
{
  return shared_ioctl(&lowtap, inode, file, cmd, arg);
} /* low_ioctl */


static int
low_read(struct inode *inode, struct file *file, char *buf, int count)
{
  struct sk_buff	*skblist[ETHTAP_MAX_READPKTS];
  unsigned		minor	= MINOR(inode->i_rdev);
  int			avail	= count - 1;	/* reserve a byte 	*/
  char			*p	= &buf[1];	/* reserved byte 	*/
  unsigned char		pkts;
  struct sk_buff	*skb;
  short			len;
  int			i, j;
  int			flags;
  tap_t		*tap = &lowtap;

  if ( minor )
    return -ENODEV;

  /* All reads are done non-blocking.  Collect as many packets as you
   * can and store them into the userspace buffer in the form
   * [<pkts>b<len>2<pkt>?<len>2<pkt>?...].  As we collect packet we
   * can either run out of slots in the skblist or run out of room in
   * the buffer.  Running out of buffer space is detected late so a
   * static pointer is used to keep track of any remaining buffer
   * between reads.  Closing the device will free the space.
   *
   * With interrupts disabled suck up as many packets as possible and
   * then with interrupts enabled fill the userspace buffer.
   */
  i = 0;			/* goto slot zero			*/
  save_flags(flags); cli();	/* -lock- 				*/
  if ( tap->lskb ) {		/* collect extra from last read 	*/
    skblist[i++] = tap->lskb;
  }
  while ( i < ETHTAP_MAX_READPKTS ) {
    tap->lskb = skb_dequeue(&tap->rxq);
    if ( tap->lskb ) {
      tap->rxsize--;
      if (avail > (tap->lskb->len + sizeof(struct ethhdr) + sizeof len)) {
	avail -= tap->lskb->len + sizeof(struct ethhdr) + sizeof len;
	skblist[i] = tap->lskb;	/* hold onto the new packet		*/
	tap->lskb  = NULL;	/* reset the static packet		*/
	i++;			/* next packet slot 			*/
      } else break;
    } else break;
  }
  restore_flags(flags);		/* -unlock- 				*/

  for (j=0; j<i; j++) {
    skb = skblist[j];

    len = skb->len + sizeof(struct ethhdr);

    memcpy_tofs(p, &len, sizeof len);
    p += sizeof len;
    memcpy_tofs(p, skb->mac.raw, len);
    p += len;
    
    kfree_skb(skb, FREE_READ);
  }
  
				/* fill in total num of pkts stored in
				   the buffer */
  pkts = i;
  memcpy_tofs(buf, &pkts, sizeof pkts);
  tap->rxread += pkts;		/* update statistics 			*/
  
#ifdef CONFIG_ETHTAP_DEBUG
  if ( j )
    TAPDEBUG("%s: read %d pkts\n", tap->name, j);
#endif  

  return pkts ? count - avail : 0;
} /* low_read */


static int
low_write(struct inode *inode, struct file *file, const char *buf, int count) 
{
  return shared_write(&lowtap, inode, file, buf, count);
} /* low_write */


static int
low_do_hard_start_xmit(void) 
{
  struct sk_buff	*skb;
  int			flags;
  int			busy = 0;
  

  save_flags(flags); cli();	/* -lock- 				*/
  skb = skb_dequeue(&lowtap.txq);
  if ( skb ) {
    lowtap.txsize--;
    busy = hard_start_xmit(skb, lld);
  }
  restore_flags(flags);		/* -unlock- 				*/

  if ( !busy ) {
    lowtap.txread++;
  }
  else {
    printk("low_do_hard_start_xmit: busy\n");
    save_flags(flags); cli();	/* -lock- 				*/
    skb_queue_head(&lowtap.txq, skb);
    lowtap.txsize++;
    restore_flags(flags);	/* -unlock- 				*/
    return 1;
  }
  
  return 0;
} /* low_do_hard_start_xmit */


/* 
 * Shared /dev/ltap, /dev/htap methods 
 */


static int
shared_open(tap_t *tap, struct inode *inode, struct file *file, int rw)
{
  unsigned	minor = MINOR(inode->i_rdev);

  if ( minor )
    return -ENODEV;

  if ( tap->mode != ETHTAP_CLOSED_MODE )
    return -EBUSY;

  tap_purge_rx(tap, ETHTAP_MAX_QSIZE, rw);
  tap->mode = ETHTAP_PASSIVE_MODE;
  return 0;
} /* shared_open */


static void
shared_release(tap_t *tap, struct inode *inode, struct file *file)
{
  unsigned	minor = MINOR(inode->i_rdev);

  if ( minor )
    return;
  
  tap->mode = ETHTAP_CLOSED_MODE;
} /* shared_release */


static int
shared_ioctl(tap_t *tap, struct inode *inode, struct file *file,
	     unsigned cmd, unsigned long arg)
{
  unsigned	minor = MINOR(inode->i_rdev);
  int		retval = 0;

  if ( minor )
    return -ENODEV;

  switch ( cmd ) {
    case ETHTAPSETMODE:		/* Set mode 				*/
      if ( (retval = verify_area(VERIFY_READ, (void *)arg, sizeof(int))) )
	return retval;

      switch ( get_user((int *)arg) ) {
	case ETHTAP_PASSIVE_MODE:
	  tap->mode = ETHTAP_PASSIVE_MODE;
	  break;
	case ETHTAP_ACTIVE_MODE:
	  tap->mode = ETHTAP_ACTIVE_MODE;
	  break;
	default:
	  retval = -EINVAL;
      }
      break;
      
    case ETHTAPGETMODE:		/* Get mode 				*/
      if ( (retval = verify_area(VERIFY_WRITE, (void *)arg, sizeof(int))) )
	return retval;
	
      put_user(tap->mode, (int *)arg);
      break;

    case ETHTAPZEROCOUNTS:	/* Reset counts 			*/
      tap->rxread  = 0;
      tap->rxwrite = 0;
      tap->rxlost  = 0;
      tap->txread  = 0;
      tap->txwrite = 0;
      tap->txlost  = 0;
      break;
      
    default:
      retval = -EINVAL;
  }

  return retval;
} /* shared_ioctl */


static int
shared_write(tap_t *tap, struct inode *inode, struct file *file,
	     const char *buf, int count) 
{
  unsigned		minor	= MINOR(inode->i_rdev);
  const char		*p	= (char *)buf;
  unsigned char		pkts;
  unsigned short	len;
  int			i;

  if ( minor )
    return -ENODEV;

  if ( tap->mode != ETHTAP_CLOSED_MODE ) {
    pkts = get_user((unsigned char *)p);
    p   += sizeof pkts;
  
    for (i=0; i < pkts; i++) {
      len = get_user((unsigned short *)p);
      p  += sizeof len;
      if ( tap->write_skb(p, len) == len)
	tap->rxwrite++;
      p  += len;
    }

    TAPDEBUG("%s: write %d pkts\n", tap->name, i);
    return count;
  }

  return -1;
} /* shared_write */



/* 
 * Private functions
 */

static void
tap_init(tap_t *tap, const char *name,
		int (*fn)(const char *, int))
{
  skb_queue_head_init(&tap->rxq);
  skb_queue_head_init(&tap->txq);
  tap->mode 	 = ETHTAP_CLOSED_MODE;
  tap->lskb	 = NULL;
  tap->rxsize	 = 0;
  tap->txsize	 = 0;
  tap->rxread 	 = 0;
  tap->txread 	 = 0;
  tap->rxwrite	 = 0;
  tap->txwrite	 = 0;
  tap->rxlost	 = 0;
  tap->txlost	 = 0;
  tap->name	 = (char *)name;
  tap->write_skb = fn;
} /* tap_init */


static void
tap_purge_rx(tap_t *tap, int num, int rw) 
{
  struct sk_buff	*skb;
  int			i = 0;

  while ( i<num && (skb = skb_dequeue(&tap->rxq)) ) {
    kfree_skb(skb, rw);
    tap->rxsize--;
    tap->rxlost++;
    i++;
  }
} /* tap_purge_rx */


static void
tap_purge_tx(tap_t *tap, int num, int rw) 
{
  struct sk_buff	*skb;
  int			i = 0;

  while ( i<num && (skb = skb_dequeue(&tap->txq)) ) {
    kfree_skb(skb, rw);
    tap->txsize--;
    tap->txlost++;
    i++;
  }
} /* tap_purge_tx */



/* ethtap.c */




